==> mcss-0.0.39.tar.gz: mcss-0.0.39/mcss/MulticompartmentGillespieQ2GD.cpp <==
// multicompartment gillespie direct method (growth and division with queue)
// class definitions
// copyright 2008, 2009, 2010 jamie twycross, jpt AT cs.nott.ac.uk
// released under GNU GPL version 3

#if HAVE_CONFIG_H
	#include <config.h>
#endif

#include "MulticompartmentGillespieQ2GD.h"

#include <limits>
#include <ecsb/ecsb.h>
#include <gsl/gsl_sf_log.h>

using namespace Simulation;
using namespace ECSB;

// constructor
MulticompartmentGillespieQ2GD::MulticompartmentGillespieQ2GD( \
		const unsigned long seed) : SimulationAlgorithm( \
		"Multicompartment Gillespie Queue 2 Growth Division", seed)
{
	waiting_times = NULL;
	compartment_reactions = NULL;
	compartment_reactions_division = NULL;
	last_volume_update = NULL;
}

// destructor
MulticompartmentGillespieQ2GD::~MulticompartmentGillespieQ2GD()
{
	if(waiting_times)
		bhpq_free(waiting_times);
	if(compartment_reactions)
		free(compartment_reactions);
	if(compartment_reactions_division)
		free(compartment_reactions_division);
	if(last_volume_update)
		free(last_volume_update);
}

// initialise algorithm
int MulticompartmentGillespieQ2GD::preprocess(void)
{
	register unsigned long i;
	register double tautmp;
	Compartment *compartment;

	// allocate memory for hash of global waiting times
	if(waiting_times)
		bhpq_free(waiting_times);
	waiting_times = bhpq_init(num_compartments);

	// allocate memory to store number of reactions performed in compartment
	if(compartment_reactions)
		free(compartment_reactions);
	if(!(compartment_reactions = (unsigned long *) \
			calloc(num_compartments, sizeof(unsigned long)))) {
		cerr << "error: couldn't allocate memory for compartment reactions\n";
		exit(EXIT_FAILURE);
	}

	// allocate memory to store number of reactions performed in compartment
	if(compartment_reactions_division)
		free(compartment_reactions_division);
	if(!(compartment_reactions_division = (unsigned long *) \
			calloc(num_compartments, sizeof(unsigned long)))) {
		cerr << "error: couldn't allocate memory for compartment " \
				"division reactions\n";
		exit(EXIT_FAILURE);
	}

	// allocate memory to store number of last volume update in compartment
	if(last_volume_update)
		free(last_volume_update);
	if(!(last_volume_update = (double *) \
			calloc(num_compartments, sizeof(double)))) {
		cerr << "error: couldn't allocate memory for compartment " \
				"volume updates\n";
		exit(EXIT_FAILURE);
	}

	// calculate tau and mu for each compartment
	for(i = 0; i < num_compartments; i++) {
		compartment = psystem->getCompartment(i);
		psystem->execute(i);
		tautmp = compartment->getTimeToNextReaction();
		// add tau to map of compartment waiting times
		if(tautmp != 0.0 || compartment->isDividing()) {
			qe.key1 = tautmp + tau0global;
			qe.key2 = (int) i;
			bhpq_insert(&qe, waiting_times);
		}
		// reset number of reactions and volume update
		compartment_reactions[i] = 0;
		compartment_reactions_division[i] = 0;
		last_volume_update[i] = 0.0;
	}

	// finish if no more reactions to execute
	if(!bhpq_peek_head(&qe, waiting_times)) {
		tau0 = mu0 = i0 = rs0 = 0;
		return 0;
	}

	// get compartment with smallest waiting time
	tau0global = qe.key1;
	i0 = (unsigned long) qe.key2;
	compartment = psystem->getCompartment(i0);
	tau0 = compartment->getTimeToNextReaction();
	mu0 = compartment->getNextReaction();
	rs0 = compartment->getNextRuleSet();

	return 1;
}

// rescale internal data structures after compartment division
void MulticompartmentGillespieQ2GD::rescale(void)
{
	// reallocate memory for hash of global waiting times
	waiting_times = bhpq_resize(waiting_times, num_compartments);

	// reallocate memory to store number of reactions performed in compartment
	if(!(compartment_reactions = (unsigned long *) realloc((void *) \
			compartment_reactions, num_compartments * sizeof(unsigned long)))) {
		cerr << "error: couldn't reallocate memory for compartment " \
				"reactions\n";
		exit(EXIT_FAILURE);
	}
	compartment_reactions[num_compartments - 1] = 0;

	// reallocate memory to store number of reactions performed in compartment
	if(!(compartment_reactions_division = (unsigned long *) realloc((void *) \
			compartment_reactions_division, num_compartments * \
			sizeof(unsigned long)))) {
		cerr << "error: couldn't reallocate memory for compartment " \
				"reactions\n";
		exit(EXIT_FAILURE);
	}
	compartment_reactions_division[num_compartments - 1] = 0;

	// reallocate memory to store number of last volume update in compartment
	if(!(last_volume_update = (double *) realloc((void *) last_volume_update, \
			num_compartments * sizeof(double)))) {
		cerr << "error: couldn't allocate memory for compartment " \
				"volume updates\n";
		exit(EXIT_FAILURE);
	}
	last_volume_update[num_compartments - 1] = time;
}

// perform one iteration of algorithm for all compartments
// using multicompartment gillespie algorithm with queue
int MulticompartmentGillespieQ2GD::execute(void)
{
	register unsigned long it;
	register double tautmp;
	register Compartment *compartment;
	register bool division;

	// update simulation time
	setTime(tau0global);

	// perform reaction in selected compartment
	compartment = psystem->getCompartment(i0);
	if((division = compartment->isDividing())) {
		divideCompartment(i0);  // division event
		if(keep_divisions)
			rescale();
	} else
		psystem->executeRule(rs0, mu0, i0);  // reaction event

	// update number of reactions executed in compartment
	if(!compartment->isVolumeConstant()) {
		compartment_reactions[i0]++;
		if(!division)
			compartment_reactions_division[i0]++;
	}

	// run gillespie to recalculate next reaction in selected compartment
	psystem->execute(i0);

	// add new waiting time to map of compartment waiting times
	tautmp = compartment->getTimeToNextReaction();
	if(tautmp != 0.0 || compartment->isDividing()) {
		qe.key1 = tautmp + tau0global;
		qe.key2 = (int) i0;
		bhpq_replace_head(&qe, waiting_times);
	} else
		// remove old waiting time from map of compartment waiting times
		bhpq_delete_head(waiting_times);

	// update target compartment if channel
	if(!division && compartment->isChannel(rs0, mu0)) {
		it = compartment->getRuleTarget(rs0, mu0);
		// run gillespie to recalculate next reaction in target compartment
		psystem->execute(it);
		// add new waiting time to map of compartment waiting times
		compartment = psystem->getCompartment(it);
		tautmp = compartment->getTimeToNextReaction();
		if(tautmp != 0.0 || compartment->isDividing()) {
			qe.key1 = tautmp + tau0global;
			qe.key2 = (int) it;
			bhpq_insert(&qe, waiting_times);
		}
	}

	// add daughter compartment to queue
	if(division && keep_divisions) {
		it = num_compartments - 1;
		// run gillespie to recalculate next reaction in target compartment
		psystem->execute(it);
		// add new waiting time to map of compartment waiting times
		compartment = psystem->getCompartment(it);
		tautmp = compartment->getTimeToNextReaction();
		if(tautmp != 0.0 || compartment->isDividing()) {
			qe.key1 = tautmp + tau0global;
			qe.key2 = (int) it;
			bhpq_insert(&qe, waiting_times);
		}
	}

	// finish if no more reactions to execute
	if(!bhpq_peek_head(&qe, waiting_times)) {
		tau0 = mu0 = i0 = rs0 = 0;
		return 0;
	}

	// get compartment with smallest waiting time
	tau0global = qe.key1;
	i0 = (unsigned long) qe.key2;
	compartment = psystem->getCompartment(i0);
	tau0 = compartment->getTimeToNextReaction();
	mu0 = compartment->getNextReaction();
	rs0 = compartment->getNextRuleSet();

	return 1;
}

// perform one iteration of algorithm for one compartment
// using gillespie direct method
void MulticompartmentGillespieQ2GD::execute(unsigned long c)
{
	register unsigned long i, mu0, rs0;
	register double a0, tau0, r1, r2, am;
	register Compartment *compartment;

	// don't run for compartments with no rules
//	if(!compartment->getNumRules())
//		return;

	compartment = psystem->getCompartment(c);

	// handle compartment growth and division if required
	if(!compartment->isVolumeConstant()) {
		if(compartment_reactions_division[c] >= division_rate) {
			if(compartment->evaluateDivisionFunction(time, tau0)) {
				// compartment has reached division volume
				// set next reaction time to division time
				tau0 = 0.0;
				mu0 = 0;
				rs0 = 0;
				compartment->setDividing(true);
				compartment_reactions_division[c] = 0;
				goto done;
			}
		}
		if(compartment_reactions[c] >= volume_rate) {
			// increase compartment volume
			compartment->advanceVolume(time, time - last_volume_update[c]);
			compartment_reactions[c] = 0;
			last_volume_update[c] = time;
		}
	}

	// calculate propensities
	a0 = 0.0;
	for(rs0 = 0; rs0 < compartment->getNumRuleSets(); rs0++)
		for(i = 0; i < compartment->getNumRules(rs0); i++)
			a0 += calculatePropensity(rs0, i, c);

	// finish if zero level
	if(a0 == 0.0) {
		compartment->setTimeToNextReaction(0.0);
		compartment->setNextReaction(0);
		compartment->setNextRuleSet(0);
		return;
	}

	// generate time to next reaction
	r1 = randomNumberPos();
	tau0 = -1.0 * gsl_sf_log(r1) / a0;

	// select next reaction
	r2 = randomNumberPos();
	am = r2 * a0;
	for(rs0 = 0; rs0 < compartment->getNumRuleSets(); rs0++)
		for(mu0 = 0; mu0 < compartment->getNumRules(rs0); mu0++) {
			am -= compartment->getPropensity(rs0, mu0);
			if(am <= 0.0)
				goto done;
		}

	done:
		compartment->setTimeToNextReaction(tau0);
		compartment->setNextReaction(mu0);
		compartment->setNextRuleSet(rs0);
}

==> mcss-1.0.1.tar.gz: mcss-1.0.1/mcss/MulticompartmentGillespieQ2GD.cpp <==
// multicompartment gillespie direct method (growth and division with queue)
// class definitions
// copyright 2008, 2009, 2010 jamie twycross, jpt AT cs.nott.ac.uk
// released under GNU GPL version 3

#if HAVE_CONFIG_H
	#include <config.h>
#endif

#include "MulticompartmentGillespieQ2GD.h"

#include <limits>
#include <ecsb/ecsb.h>
#include <gsl/gsl_sf_log.h>

using namespace Simulation;
using namespace ECSB;

// constructor
MulticompartmentGillespieQ2GD::MulticompartmentGillespieQ2GD( \
		const unsigned long seed) : SimulationAlgorithm( \
		"Multicompartment Gillespie Queue 2 Growth Division", seed)
{
	waiting_times = NULL;
	compartment_reactions = NULL;
	compartment_reactions_division = NULL;
	last_volume_update = NULL;
}

// destructor
MulticompartmentGillespieQ2GD::~MulticompartmentGillespieQ2GD()
{
	if(waiting_times)
		bhpq_free(waiting_times);
	if(compartment_reactions)
		free(compartment_reactions);
	if(compartment_reactions_division)
		free(compartment_reactions_division);
	if(last_volume_update)
		free(last_volume_update);
}

// initialise algorithm
int MulticompartmentGillespieQ2GD::preprocess(void)
{
	register unsigned long i;
	register double tautmp;
	Compartment *compartment;

	// allocate memory for hash of global waiting times
	if(waiting_times)
		bhpq_free(waiting_times);
	waiting_times = bhpq_init(num_compartments);

	// allocate memory to store number of reactions performed in compartment
	if(compartment_reactions)
		free(compartment_reactions);
	if(!(compartment_reactions = (unsigned long *) \
			calloc(num_compartments, sizeof(unsigned long)))) {
		cerr << "error: couldn't allocate memory for compartment reactions\n";
		exit(EXIT_FAILURE);
	}

	// allocate memory to store number of reactions performed in compartment
	if(compartment_reactions_division)
		free(compartment_reactions_division);
	if(!(compartment_reactions_division = (unsigned long *) \
			calloc(num_compartments, sizeof(unsigned long)))) {
		cerr << "error: couldn't allocate memory for compartment " \
				"division reactions\n";
		exit(EXIT_FAILURE);
	}

	// allocate memory to store number of last volume update in compartment
	if(last_volume_update)
		free(last_volume_update);
	if(!(last_volume_update = (double *) \
			calloc(num_compartments, sizeof(double)))) {
		cerr << "error: couldn't allocate memory for compartment " \
				"volume updates\n";
		exit(EXIT_FAILURE);
	}

	// calculate tau and mu for each compartment
	for(i = 0; i < num_compartments; i++) {
		compartment = psystem->getCompartment(i);
		psystem->execute(i);
		tautmp = compartment->getTimeToNextReaction();
		// add tau to map of compartment waiting times
		if(tautmp != 0.0 || compartment->isDividing()) {
			qe.key1 = tautmp + tau0global;
			qe.key2 = (int) i;
			bhpq_insert(&qe, waiting_times);
		}
		// reset number of reactions and volume update
		compartment_reactions[i] = 0;
		compartment_reactions_division[i] = 0;
		last_volume_update[i] = 0.0;
	}

	// finish if no more reactions to execute
	if(!bhpq_peek_head(&qe, waiting_times)) {
		tau0 = mu0 = i0 = rs0 = 0;
		return 0;
	}

	// get compartment with smallest waiting time
	tau0global = qe.key1;
	i0 = (unsigned long) qe.key2;
	compartment = psystem->getCompartment(i0);
	tau0 = compartment->getTimeToNextReaction();
	mu0 = compartment->getNextReaction();
	rs0 = compartment->getNextRuleSet();

	return 1;
}

// rescale internal data structures after compartment division
void MulticompartmentGillespieQ2GD::rescale(void)
{
	// reallocate memory for hash of global waiting times
	waiting_times = bhpq_resize(waiting_times, num_compartments);

	// reallocate memory to store number of reactions performed in compartment
	if(!(compartment_reactions = (unsigned long *) realloc((void *) \
			compartment_reactions, num_compartments * sizeof(unsigned long)))) {
		cerr << "error: couldn't reallocate memory for compartment " \
				"reactions\n";
		exit(EXIT_FAILURE);
	}
	compartment_reactions[num_compartments - 1] = 0;

	// reallocate memory to store number of reactions performed in compartment
	if(!(compartment_reactions_division = (unsigned long *) realloc((void *) \
			compartment_reactions_division, num_compartments * \
			sizeof(unsigned long)))) {
		cerr << "error: couldn't reallocate memory for compartment " \
				"reactions\n";
		exit(EXIT_FAILURE);
	}
	compartment_reactions_division[num_compartments - 1] = 0;

	// reallocate memory to store number of last volume update in compartment
	if(!(last_volume_update = (double *) realloc((void *) last_volume_update, \
			num_compartments * sizeof(double)))) {
		cerr << "error: couldn't allocate memory for compartment " \
				"volume updates\n";
		exit(EXIT_FAILURE);
	}
	last_volume_update[num_compartments - 1] = time;
}

// perform one iteration of algorithm for all compartments
// using multicompartment gillespie algorithm with queue
int MulticompartmentGillespieQ2GD::execute(void)
{
	register unsigned long it;
	register double tautmp;
	register Compartment *compartment;
	register bool division;

	// update simulation time
	setTime(tau0global);

	// perform reaction in selected compartment
	compartment = psystem->getCompartment(i0);
	if((division = compartment->isDividing())) {
		divideCompartment(i0);  // division event
		if(keep_divisions)
			rescale();
	} else
		psystem->executeRule(rs0, mu0, i0);  // reaction event

	// update number of reactions executed in compartment
	if(!compartment->isVolumeConstant()) {
		compartment_reactions[i0]++;
		if(!division)
			compartment_reactions_division[i0]++;
	}

	// run gillespie to recalculate next reaction in selected compartment
	psystem->execute(i0);

	// add new waiting time to map of compartment waiting times
	tautmp = compartment->getTimeToNextReaction();
	if(tautmp != 0.0 || compartment->isDividing()) {
		qe.key1 = tautmp + tau0global;
		qe.key2 = (int) i0;
		bhpq_replace_head(&qe, waiting_times);
	} else
		// remove old waiting time from map of compartment waiting times
		bhpq_delete_head(waiting_times);

	// update target compartment if channel
	if(!division && compartment->isChannel(rs0, mu0)) {
		it = compartment->getRuleTarget(rs0, mu0);
		// run gillespie to recalculate next reaction in target compartment
		psystem->execute(it);
		// add new waiting time to map of compartment waiting times
		compartment = psystem->getCompartment(it);
		tautmp = compartment->getTimeToNextReaction();
		if(tautmp != 0.0 || compartment->isDividing()) {
			qe.key1 = tautmp + tau0global;
			qe.key2 = (int) it;
			bhpq_insert(&qe, waiting_times);
		}
	}

	// add daughter compartment to queue
	if(division && keep_divisions) {
		it = num_compartments - 1;
		// run gillespie to recalculate next reaction in target compartment
		psystem->execute(it);
		// add new waiting time to map of compartment waiting times
		compartment = psystem->getCompartment(it);
		tautmp = compartment->getTimeToNextReaction();
		if(tautmp != 0.0 || compartment->isDividing()) {
			qe.key1 = tautmp + tau0global;
			qe.key2 = (int) it;
			bhpq_insert(&qe, waiting_times);
		}
	}

	// finish if no more reactions to execute
	if(!bhpq_peek_head(&qe, waiting_times)) {
		tau0 = mu0 = i0 = rs0 = 0;
		return 0;
	}

	// get compartment with smallest waiting time
	tau0global = qe.key1;
	i0 = (unsigned long) qe.key2;
	compartment = psystem->getCompartment(i0);
	tau0 = compartment->getTimeToNextReaction();
	mu0 = compartment->getNextReaction();
	rs0 = compartment->getNextRuleSet();

	return 1;
}

// perform one iteration of algorithm for one compartment
// using gillespie direct method
void MulticompartmentGillespieQ2GD::execute(unsigned long c)
{
	register unsigned long i, mu0, rs0;
	register double a0, tau0, r1, r2, am;
	register Compartment *compartment;

	// don't run for compartments with no rules
//	if(!compartment->getNumRules())
//		return;

	compartment = psystem->getCompartment(c);

	// handle compartment growth and division if required
	if(!compartment->isVolumeConstant()) {
		if(compartment_reactions_division[c] >= division_rate) {
			if(compartment->evaluateDivisionFunction(time, tau0)) {
				// compartment has reached division volume
				// set next reaction time to division time
				tau0 = 0.0;
				mu0 = 0;
				rs0 = 0;
				compartment->setDividing(true);
				compartment_reactions_division[c] = 0;
				goto done;
			}
		}
		if(compartment_reactions[c] >= volume_rate) {
			// increase compartment volume
			compartment->advanceVolume(time, time - last_volume_update[c]);
			compartment_reactions[c] = 0;
			last_volume_update[c] = time;
		}
	}

	// calculate propensities
	a0 = 0.0;
	for(rs0 = 0; rs0 < compartment->getNumRuleSets(); rs0++)
		for(i = 0; i < compartment->getNumRules(rs0); i++)
			a0 += calculatePropensity(rs0, i, c);

	// finish if zero level
	if(a0 == 0.0) {
		compartment->setTimeToNextReaction(0.0);
		compartment->setNextReaction(0);
		compartment->setNextRuleSet(0);
		return;
	}

	// generate time to next reaction
	r1 = rng->randomNumberPos();
	tau0 = -1.0 * gsl_sf_log(r1) / a0;

	// select next reaction
	r2 = rng->randomNumberPos();
	am = r2 * a0;
	for(rs0 = 0; rs0 < compartment->getNumRuleSets(); rs0++)
		for(mu0 = 0; mu0 < compartment->getNumRules(rs0); mu0++) {
			am -= compartment->getPropensity(rs0, mu0);
			if(am <= 0.0)
				goto done;
		}

	done:
		compartment->setTimeToNextReaction(tau0);
		compartment->setNextReaction(mu0);
		compartment->setNextRuleSet(rs0);
}
